#include <linux/version.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
#include <linux/device.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
# define usb_match_id(dev, intf, table) usb_match_id(intf, table)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
# define return_probe_ok(priv_data) return priv_data
# define return_probe_fail(retval)  return NULL
# define usb_set_intfdata(intf, priv_data)
#else
# define return_probe_ok(priv_data) return 0
# define return_probe_fail(retval) return retval
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
extern inline void module_put(struct module *mod)
{
	BUG_ON(mod != THIS_MODULE);
	MOD_DEC_USE_COUNT;
}
extern inline int try_module_get(struct module *mod)
{
	BUG_ON(mod != THIS_MODULE);
	MOD_INC_USE_COUNT;
	return 1;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
   /* Take not of the inline functions to hack around modversions */

# ifdef usb_submit_urb
   extern inline int usb_submit_urb_real(struct urb *urb)
   {
	return usb_submit_urb(urb);
   }
#  undef usb_submit_urb
#  define usb_submit_urb(urb, mem_flags) usb_submit_urb_real(urb)
# else
#  define usb_submit_urb(urb, mem_flags) usb_submit_urb(urb)
# endif

# ifdef usb_alloc_urb
   extern inline struct urb *usb_alloc_urb_real(int iso_packets)
   {
	return usb_alloc_urb(iso_packets);
   }
#  undef usb_alloc_urb
#  define usb_alloc_urb(iso_packets, mem_flags) usb_alloc_urb_real(iso_packets)
# else
#  define usb_alloc_urb(iso_packets, mem_flags) usb_alloc_urb(iso_packets)
# endif
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,69)
static inline
void usb_init_urb(struct urb *urb)
{
	memset(urb, 0, sizeof(*urb));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) /* tune me! */
	urb->count = (atomic_t)ATOMIC_INIT(1);
#endif
	spin_lock_init(&urb->lock);
}
#endif


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#define bridge_bulk_in_callback(urb, pt_regs) \
		bridge_bulk_in_callback(urb)
#define bridge_request_out_callback(urb, pt_regs) \
		bridge_request_out_callback(urb)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#define URB_ASYNC_UNLINK USB_ASYNC_UNLINK
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	
# define get_iface_desc(iface, alt_n) (&(iface)->altsetting[alt_n])

# define get_iface_ep_desc(iface, alt_n, n) \
		(&(iface)->altsetting[alt_n].endpoint[n])

#else

# define get_iface_desc(iface, alt_n) (&(iface)->altsetting[alt_n].desc)

# define get_iface_ep_desc(iface, alt_n, n) \
		(&(iface)->altsetting[alt_n].endpoint[n].desc)

#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#define fw_dev_param(dev) (dev->net_dev->name)
#else
#define fw_dev_param(dev) (&dev->udev->dev)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#define WARN_ON BUG_ON
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20))
static inline void list_move_tail(struct list_head *list,
				  struct list_head *head)
{
        __list_del(list->prev, list->next);
        list_add_tail(list, head);
}
#endif
