/*
 * Hermes AP firmware extractor for Windows drivers (c) 2003 by Mark Smith
 * This may be distributed freely under the GPL v2 so long as this copyright
 * notice is included.
*/

/*
 * Hermes Ap Usb Firmware Extractor for windows drivers (c)2003 by Alessandro Faillace
 * based on the work of Mark Smith.
 * This may be distributed freely under the GPL v2 so long as this copyright
 * notice is included.
 */

#include <stdio.h>     /* for printf */
#include <stdlib.h>    /* for exit */
#include <memory.h>
#include <string.h>
#include <getopt.h>
#include <asm/byteorder.h>

#define false   0
#define true    (!0)

#define VERSION  "0.3"

struct _segarray {
	unsigned int offset;
	unsigned int size;
	unsigned char *data;
};

struct _plugarray {
	unsigned int code;
	unsigned int targ_off;
	unsigned int length;
};

struct _usb_fw {
	unsigned int fw_start_offset;
	unsigned int fw_end_offset;
	unsigned int variant_offset;
	unsigned int size;
	unsigned char code[0x1b40];
};

struct firmwareblock {
	struct _segarray *segarray;
	unsigned int halfentry;
	struct _plugarray *plugarray;
} *firmware;

struct _windrv {
	unsigned int flen;
	unsigned char data[0];
};

void usage(char *argv[]);
void credits();
void checklimits(unsigned int sigpos, unsigned int Llimit,unsigned int Ulimit);
struct _usb_fw *extractusbfirmware(struct _windrv *windrv);
void savefirmware(struct _windrv *wdata, unsigned int signature, unsigned char hexchar);
int  memfinder(unsigned char *data, unsigned int flen, unsigned char search[], unsigned int searchlen);
void usbfwdumper(unsigned int firmstart,unsigned int firmend, unsigned char *data,char *filename,int style);
int backfinder (unsigned char *offset,unsigned char *data,unsigned char search[], unsigned int searchlen);
void variantcheck (struct _usb_fw *usbfw);
void usb_ver004dumper (struct _usb_fw *usbfw,char *filename);
void usb_ver005dumper (struct _usb_fw *usbfw,char *filename);
void usb_ver006dumper (struct _usb_fw *usbfw,char *filename);
void usb_bindumper (struct _usb_fw *usbfw,char *filename);
long big2little(long inval);
struct _windrv *readdrvfile (char *filename);

int main(int argc, char *argv[])
{
	int c;
	int verbose=0;
	struct _windrv *data;
	struct _usb_fw *usb_fw;

	while ((c = getopt(argc, argv, "VD:v:t:s:b:5:4:6:")) != EOF)
	{
		switch (c) {
			case 'V':
				usage (&argv[0]);
				break;
			case 'v':
				verbose=1;
				break;
			case 'b':
				usb_fw = extractusbfirmware(readdrvfile(argv[2]));
				variantcheck (usb_fw);
				usb_bindumper(usb_fw,"orinoco_usb_fw.bin");
				return 0;
			case '6':
				usb_fw = extractusbfirmware(readdrvfile(argv[2]));
				variantcheck (usb_fw);
				usb_ver006dumper(usb_fw,"orinoco_usb_fw.h");
				return 0;
			case '5':
				usb_fw = extractusbfirmware(readdrvfile(argv[2]));
				variantcheck (usb_fw);
				usb_ver005dumper(usb_fw,"orinoco_usb_fw.h");
				return 0;
			case '4':
				usb_fw = extractusbfirmware(readdrvfile(argv[2]));
				variantcheck (usb_fw);
				usb_ver004dumper(usb_fw,"orinoco_usb_fw.h");
				return 0;
			case 't':
				data=readdrvfile(argv[2]);
				savefirmware(data, 0xfbfe4461, 'T');
				return 0;
			case 's':
				data=readdrvfile(argv[2]);
				savefirmware(data, 0x63fc600f, 'R');
				return 0;
			default:
				usage (&argv[0]);
				break;
		}
	}
	usage (&argv[0]);
	return 0;
}

struct _windrv *readdrvfile (char *filename)
	{
	FILE *input;
	unsigned int flen;
	struct _windrv *windrv;

	if ((input = fopen(filename, "rb")) == NULL) {
		printf("Unable to open %s, aborting.\n", filename);
		exit(1);
	}
	// Get file length
	fseek(input, 0L, SEEK_END);
	flen = ftell(input);
	printf("File %s length %u (0x%08x)\n", filename, flen, flen);
	
	// Rewind file pointer
	fseek(input, 0L, SEEK_SET);
	
	// Allocate memory and load the file
	windrv = malloc (sizeof *windrv + flen);
	windrv->flen=flen;
	fread(windrv->data, 1, flen, input);
	fclose(input);
	printf("Memory allocated and file read OK\n");
	return windrv;
}


void usage(char *argv[]){
	credits();
        printf("\n");
        printf("Usage: %s [options] wl???.sys\n", argv[0]);
        printf("\nUsb bootloader firmware extraction\n");
        printf("\t -4 extract usb firmware in include format 0.0.4 Version\n");
        printf("\t -5 extract usb firmware in include format 0.0.5 Version\n");
        printf("\t -6 extract usb firmware in include format 0.0.6 Version\n");
        printf("\t -b extract usb firmware in binary  format\n");
        printf("\nTertiary and station firmware extraction\n");
        printf("\t -s extract station firmware in hfw format\n");
        printf("\t -t extract tertiary  firmware in hfw format\n");
        printf("\nMisc\n");
        printf("\t -h, --help this help\n");
        printf("\t -V, --version version number and credits\n");
	exit(0);
}

void credits(){
	printf("Lucent Firmware Extractor v1.0 alpha\n(c) 2003 Mark Smith (username 'Mark' on HermesAP board)\n");
	printf("Orinco usb firmware extractor  v%s (c) 2003 Alessandro Faillace\n",VERSION);
	printf("Distributed under GPL v2 so long as this copyright notice is included.\n");
}

void checklimits(unsigned int sigpos, unsigned int Llimit,unsigned int Ulimit)
{
        printf("Checking distance ... 0x%08x...", sigpos);
        if (sigpos > Ulimit) {
                printf("to far away\n");
        } else if (sigpos < Llimit) {
                printf("to close\n");
        } else {
                printf("ok\n");
        }
}

struct _usb_fw *extractusbfirmware(struct _windrv *windrv)
{
        unsigned char *offset, *page1start;
        unsigned char hexstring[16];
        unsigned int i; 
        unsigned char Usb_Clinet_sig[] = {0x55,0x00,0x53,0x00,0x42,0x00,0x20,0x00,0x43,0x00,0x6c,0x00,0x69,0x00,0x65,0x00,0x6e,0x00,0x74};
        unsigned char Softronics_sig[] = {0x53,0x00,0x6F,0x00,0x66,0x00,0x74,0x00,0x72,0x00,0x6F,0x00,0x6E,0x00,0x69,0x00,0x63,0x00,0x73};
        unsigned char sig2[] = {0x00,0x00,0x00,0x40};
        unsigned char sig3[] = {0x00,0x00,0x00,0x02};
        unsigned char sig4[] = {0xAA,0xAA,0xAA,0xAA};
        unsigned int  firmstart,firmend;
        int sig1off,sig2off;
        int sig2_sig1dist;
        int fwoffset;
	unsigned char *data = windrv->data;
	struct _usb_fw *usb_fw;
	printf("Lucent ORiNOCO Usb Firmware Extractor v %s \n(c) 2003 Alessandro Faillace\n" ,VERSION);
//-------------------------------------------------------------------------------------------------
// Get Start Signature
// Search for -->USB Client<-- (unicoded) then look som 0x97 or so bytes for 0x00000040
// if i find the second signatuture i then look for the startpoint by searching 0x00000020
//-------------------------------------------------------------------------------------------------

	fwoffset=memfinder(data,windrv->flen,Usb_Clinet_sig,19);
        offset=fwoffset+data;

        strncpy(hexstring, offset, 13);
        hexstring[13] = 0;
        printf("Found usb firmware at file offset 0x%08x\n", offset-data);
// Get The second  Signature to bee shure
        printf("Searching for firmware second signature ...");
        sig1off=offset-data;

	fwoffset = backfinder (offset,data,sig2,4);
        offset=fwoffset+data;

        sig2off=page1start - data;
        sig2_sig1dist=sig1off-sig2off;
        checklimits(sig2_sig1dist,0x50,0x110);


// Get The Staring point of the firmware basicly search for 0x00000002
        printf("Searching for firmware start signature ...");

	fwoffset = backfinder (offset,data,sig3,4);
        offset=fwoffset+data;

        sig2off=offset - data;
        sig2_sig1dist=sig1off-sig2off;

        checklimits(sig2_sig1dist,0x180,0x210);
        firmstart=sig2off+4;
//-------------------------------------------------------------------------------------------------
// Get End Signature
// search for Softronics and then look for 0xAAAAAA which is the end
//-------------------------------------------------------------------------------------------------
	fwoffset=memfinder(data,windrv->flen,Softronics_sig,19);
        offset=fwoffset+data;

        printf("Found usb firmware end at file offset 0x%08x\n", offset-data);


	fwoffset = backfinder (offset,data,sig4,4);
        offset=fwoffset+data;
        firmend=fwoffset+4;

	usb_fw = malloc (sizeof *usb_fw + 1 );
//
// Filling the structure 	
//
	usb_fw->fw_start_offset=firmstart;
	usb_fw->fw_end_offset=firmend;
	usb_fw->size=usb_fw->fw_end_offset-usb_fw->fw_start_offset+1;

        printf("Firmware Start   :0x%08x...\n", usb_fw->fw_start_offset);
        printf("Firmware End     :0x%08x...\n", usb_fw->fw_end_offset);
        printf("Firmware lenght  :0x%04X",usb_fw->size);

        if (usb_fw->size==0x1B40){
                printf ("...Good lenght\n");
        } else {
                printf ("\a...Bad lenght must be 0x1B40 bytes\n");
                exit(-2);
		return 0;
        }
	offset = data;
        for (i = 0 ; i < usb_fw->fw_start_offset ;i++ )
                offset++;
		
        for (i=0 ; i< usb_fw->size;i++){
		usb_fw->code[i]=*(unsigned char *)offset;
	//	usb_fw->code[i]=*offset;
		offset++;
	}

	return usb_fw;

}

void variantcheck (struct _usb_fw *usb_fw){
	int i;
        unsigned char variant[] = {0x02,0x03,0x58};

	printf("Size  0x%08x...\n", usb_fw->size);
	for (i=0;i<usb_fw->size;i++){
		if (memcmp(variant, &usb_fw->code[i] , 3) == 0){
			printf("Found variant information\n");
			usb_fw->variant_offset = i+3;
		}
	}
}

int memfinder(unsigned char *data, unsigned int flen, unsigned char search[], unsigned int searchlen){
	unsigned char *offset, *hexstringoffs, *end;
	unsigned int found;
	offset = data;
	end = (unsigned char *)((unsigned int)data + flen);
	for (found = false; found == false && offset != NULL; ) {
		offset = memchr(offset, search[0], (unsigned int)(end - offset));

		if (offset != NULL) {
			if (memcmp(search, offset , 19) == 0) {
				hexstringoffs = offset;
				found = true;
				return offset - data;
			} else
				offset++;
		}
	}

        if (offset == NULL) {
                printf("firmware not found!\n");
                exit(-1);
		return 0;
        }
	return 0;
}

int backfinder (unsigned char *offset,unsigned char *data,unsigned char search[], unsigned int searchlen){
	unsigned int found;
        for (found = false; offset > data && found == false; offset--)
		if (memcmp(search, offset , searchlen) == 0){ 
                        found = true;
		}

        if (found == false) {
                printf("Signature not found!\n");
                exit(-2);
        } else {
                printf("found it at file offset 0x%08x\n", offset - data);
		return offset - data;
        }
}

void usb_ver004dumper (struct _usb_fw *usb_fw,char *filename)
{
	int i,l,j,counter;
	FILE *output;
	printf("\n\n------------ [DUMPING] ------------\n");
	printf("Dumping to %s...\n", filename);
	if ((output = fopen(filename, "wb")) == NULL) {
		printf("Unable to open %s, aborting.\n", filename);
		exit(-5);
	}
        fprintf(output,"/*\nFirmware include file for orinoco_usb version 0.0.4\nGenerator ver %s\n*/\n",VERSION);
        fprintf(output,"static struct ihx_rec firmware[] = {\n\t");
        fprintf(output,"{ 16, 0x%04X, 0, {\n\t\t",0);
	j=0; l=0;counter=0;
	// variant patch
	usb_fw->code[usb_fw->variant_offset]=0x00;
        for (i=0 ; i< usb_fw->size;i++){
		fprintf (output,"0x%02X",usb_fw->code[i] & 0x000000ff);
		j++;l++;

		if (l >= 16){
			fprintf(output,"\n\t\t},\n\t},"); l=0;
			counter=counter+16;
		} else {
			fprintf(output,",");
		}
		if (j >= 8){ fprintf(output,"\n\t\t"); j=0; }
		if (i == usb_fw->size-1){
			fprintf(output,"\n\t{ 00, 0x%04X, 1, {\n\t\t0\n\t\t},\n\t},\n};",0);
		} else {
			if (l == 0){ fprintf(output,"\n\t{ 16, 0x%04X, 0, {\n\t\t",counter); }
		}
	}
	fputc('\n', output);
	fclose(output);
						
			
}
void usb_ver005dumper (struct _usb_fw *usb_fw,char *filename)
{
	int i,j;
	FILE *output;
	printf("\n\n------------ [DUMPING] ------------\n");
	printf("Dumping to %s...\n", filename);
	if ((output = fopen(filename, "wb")) == NULL) {
		printf("Unable to open %s, aborting.\n", filename);
		exit(-5);
	}
	fprintf(output,"/*\nFirmware include file for orinoco_usb version 0.0.5\nGenerator ver %s\n*/\n",VERSION);
	fprintf(output,"static struct ez_usb_fw firmware = {\n\t.size = 0x%02X,\n\t.code = {\n\t\t",usb_fw->size);
	j=0; 
        for (i=0 ; i< usb_fw->size;i++){
		fprintf (output,"0x%02X,",usb_fw->code[i] & 0x000000ff);
		j++;
		if (j >= 8){ fprintf(output,"\n\t\t"); j=0; }
	}
	fprintf(output,"\n\t},\n};");
	fputc('\n', output);
	fclose(output);

}
void usb_ver006dumper (struct _usb_fw *usb_fw,char *filename)
{
	int i,j;
	FILE *output;
	printf("\n\n------------ [DUMPING] ------------\n");
	printf("Dumping to %s...\n", filename);
	if ((output = fopen(filename, "wb")) == NULL) {
		printf("Unable to open %s, aborting.\n", filename);
		exit(-5);
	}
	fprintf(output,"/*\nFirmware include file for orinoco_usb version 0.0.6\nGenerator ver %s\n*/\n",VERSION);
	fprintf(output,"static u8 inkernel_firmware_code[EZ_USB_FW_SIZE] = {\n\t\t");
	j=0; 
        for (i=0 ; i< usb_fw->size;i++){
		fprintf (output,"0x%02X,",usb_fw->code[i] & 0x000000ff);
		j++;
		if (j >= 8){ fprintf(output,"\n\t\t"); j=0; }
	}
	fprintf(output,"\n};\n");
	fprintf(output,"static struct ez_usb_fw firmware = {\n\t\t.size = 0x%04X,\n\t\t.code = inkernel_firmware_code,\n};",usb_fw->size);
	fputc('\n', output);
	fclose(output);

}
void usb_bindumper (struct _usb_fw *usb_fw,char *filename)
{
	int i;
	FILE *output;
	printf("\n\n------------ [DUMPING] ------------\n");
	printf("Dumping to %s...\n", filename);
	if ((output = fopen(filename, "wb")) == NULL) {
		printf("Unable to open %s, aborting.\n", filename);
		exit(-5);
	}
        for (i=0 ; i< usb_fw->size;i++){
		fprintf (output,"%c",usb_fw->code[i] & 0x000000ff);
	}
	fclose(output);
}

long big2little(long inval)
{
	return __swab32(inval);

/* It will be cleaner to use: 
   cpuval = __be32_to_cpu(beval);

   And friends:
      __cpu_to_le64
      __le64_to_cpu
      __cpu_to_le32
      __le32_to_cpu
      __cpu_to_le16
      __le16_to_cpu
      __cpu_to_be64
      __be64_to_cpu
      __cpu_to_be32
      __be32_to_cpu
      __cpu_to_be16
      __be16_to_cpu

   With this, you can forget what machine you are running on and just
   make statements about the data you are treating.

*/
}














//------------------------------- [hfwget] ------------------------------------
//Mark's stuff
//-----------------------------------------------------------------------------
void savefirmware(struct _windrv *windrv, unsigned int signature, unsigned char hexchar)
{
        FILE *output;
        unsigned char *offset, *end, *hexstringoffs, *page1start, *peheader, *data;
        unsigned char hexstring[16];
        unsigned int i, j, found, imagebase, vfwoffs, fwblock;

	// new strucutres compatibility
	unsigned int flen =windrv->flen;
	data = windrv->data;

	printf("Lucent Firmware Extractor v1.0 alpha\n(c) 2003 Mark Smith (username 'Mark' on HermesAP board)\n");

        // Find the ?1XXYYZZ.HEX string
        offset = data;
        end = (unsigned char *)((unsigned int)data + flen);
        for (found = false; found == false && offset != NULL; ) {
                offset = memchr(offset, hexchar, (unsigned int)(end - offset));

                if (offset != NULL) {
                        if (memcmp(".HEX", offset + 8, 4) == 0) {
                                hexstringoffs = offset;
                                found = true;
                        } else
                                offset++;
                }
        }

        if (offset == NULL) {
                printf("%c-firmware not found!\n", hexchar);
                exit(-1);
        }

        strncpy(hexstring, offset, 13);
        hexstring[13] = 0;

        printf("Found firmware %s at file offset 0x%08x\n", hexstring, offset-data);
        printf("Searching for firmware start signature 0x%08x...\n", signature);

        // Really should use a mask here, but its not necessary for the moment.
        offset -= 4;
        for (found = false; offset > data && found == false; offset--)
                if (*(unsigned int *)offset == signature)
                        found = true;

        if (found == false) {
                printf("Signature not found!\n");
                exit(-2);
        } else {
                page1start = offset;

                printf("Found at file offset 0x%08x\n", page1start - data);
        }

        peheader = (unsigned char *)(*(unsigned int *)(data + 0x3c));
        printf("Reading DOS driver header...\nPE header located at file offset %p\n", peheader);
        imagebase = (unsigned int)(*(unsigned int *)((unsigned int)peheader + (unsigned int)data + 0x34));
        vfwoffs = (unsigned int)((unsigned int)page1start - (unsigned int)data + imagebase);
        printf("PE imagebase is 0x%08x, therefore virtual offset of firmware is 0x%08x\n",
                imagebase, vfwoffs);

        printf("Now searching for driver's tertiary firmware table...\n");
        offset = end;
        vfwoffs -= 3; // kludge for driver firmware structure: one redundant dword at start.

        for (found = false; offset > data && found == false; offset--)
                if (*(unsigned int *)offset == vfwoffs)
                        found = true;

        fwblock = (unsigned int)offset - 7 + imagebase - (unsigned int)data;
        if (found == false) {
                printf("Tertiary table not found - contact Mark!\n");
                exit(-3);
        } else {
                printf("Found at virtual offset 0x%08x\n", fwblock);
        }

        printf("Finding main firmware table....\n");
        offset = end;
        for (found = false; offset > data && found == false; offset--)
                if (*(unsigned int *)offset == fwblock)
                        found = true;
		

        firmware = (struct firmwareblock *)++offset;

        if (found == false) {
                printf("Main table not found - contact Mark!\n");
                exit(-4);
        } else {
                printf("Found at file offset 0x%08x\n", offset - data);
        }


        printf("Entry point at 0x%08x\n", firmware->halfentry * 2);
        (unsigned int)firmware->segarray += (unsigned int)data - imagebase;


        for (i = 0; i < 4 && firmware->segarray[i].offset != 0; i++) {
                (unsigned int)firmware->segarray[i].data -= imagebase - 4;
                printf("Segment: %d File offs: %p Target mem: 0x%08x Length 0x%08x\n",
                        i,
                        firmware->segarray[i].data,
                        firmware->segarray[i].offset,
                        firmware->segarray[i].size);
        }
        printf("Plugrecords at file offset 0x%08x\n", (unsigned int)firmware->plugarray - imagebase);

        hexstring[9] = 0;
        strcat(hexstring, "hfw");

        printf("Dumping to %s...\n", hexstring);
        if ((output = fopen(hexstring, "wb")) == NULL) {
                printf("Unable to open %s, aborting.\n", hexstring);
                exit(-5);
        }

        (unsigned int)firmware->plugarray += (unsigned int)data - imagebase;

        fprintf(output, "HFW1\nENTRY %08X\n", firmware->halfentry * 2);
        for (i = 0; firmware->plugarray[i].code != 0; i++) {
                fprintf(output, "PLUG %08X %08X %08X\n",
                        firmware->plugarray[i].code,
                        firmware->plugarray[i].targ_off,
                        firmware->plugarray[i].length);
        }

        for (i = 0; firmware->segarray[i].offset != 0; i++) {
                (unsigned int)firmware->segarray[i].data += (unsigned int)data;
                if (i != 0)
                        fprintf(output, "\n");
                fprintf(output, "SEG %08X %08X %08X",
                        firmware->segarray[i].offset,
                        firmware->segarray[i].size,
                        0);

                for (j = 0; j < firmware->segarray[i].size; j += 2) {
                        if ((j % 16) == 0)
                                fprintf(output, "\nDATA");

                        fprintf(output, " %02X%02X",
                                firmware->segarray[i].data[j],
                                firmware->segarray[i].data[j + 1]);
                }
        }

        fputc('\n', output);
        fclose(output);
        printf("Dump of %s complete.\n", hexstring);
}
